#!/bin/sh
source /etc/env

MIN_PART0_SIZE=128
MIN_PART1_SIZE=8192
MIN_PART2_SIZE=65536
MIN_PART3_SIZE=131072
MIN_FLASH_SIZE=$(( $MIN_PART0_SIZE + $MIN_PART1_SIZE  + $MIN_PART2_SIZE + $MIN_PART3_SIZE ))

ROOTFS="rootfs.squashfs"
OPTFS="optfs"
SUBFS="subfs"
DATAFS="varfs.tgz"
REMOVELIST="remove.list"
FIXENV="fixenv"

KERNEL_VOL="uImage"
RECOVERY_VOL="uImage.recovery"
FPGADATA_BB_VOL="smb_com_top.bit"
FPGADATA_RF_VOL="smb_mod.bit"
BOOTPIC_UBOOT_VOL="bootpic_u.bmp"
BOOTPIC_LINUX_VOL="bootpic.bmp"
U_BOOT_VOL="u-boot.bin"
U_LOAD="u-load_01.bin"
VERSIONINFO="VersionInfo.txt"

FLASHDEV="mtdblock"
STEPS=3
UPDATE_TITLE="Update Step "
ZIPFILE="/mnt/$2"
ZIPERR=/tmp/ziperr
CHKMSG="Checking update file..."
SAVE="/tmp/save.tar"
RECOVERY="/recovery/recovery.rsu"

#/bin/sh </dev/console >/dev/console 2>/dev/console &

center()
{
echo -geometry ${1}x${2}+$(( ( ${xsize} - ${1} ) / 2))+$(( ( ${ysize} - ${2} ) / 2))
}

title()
{
echo "${UPDATE_TITLE}${1}/${STEPS}"
}

panic()
{
	qdialog $(center 240 100) -title "Update" -fn "${qtoolfont}" -c -1 "&Reboot" -o "$1"

	umount /rw
	umount /mnt

	reboot -f
}

panic_space()
{
	[ $# -ne 0 ] && rm -rf "$*"
	panic "write error or insufficent space on storage!" 
}

file_length()
{
stat -c '%s' ${1} 2>/dev/null || echo -n -1
}
 
check_ziperr()
{
if [ $(file_length ${ZIPERR}) -ne 0 ] ; then
	cat ${ZIPERR}
	panic "Update file corrupted!"
fi
}

zip_length()
{
echo " $(unzip -l ${ZIPFILE} "${1}" 2>${ZIPERR} | grep "\<${1}$")" | tr -s ' ' | cut -f2 -d' '
check_ziperr
}

zip_all_length()
{
unzip -l ${ZIPFILE} 2>${ZIPERR} | tail -n 1 | tr -s ' ' | cut -f2 -d' '
check_ziperr
}

zip_read()
{
unzip -p ${ZIPFILE} "${1}" 2>${ZIPERR}
check_ziperr
}

md5sum_cmp()
{
[ ! -e ${1} ] && return 1
[ $(file_length ${1}) -ne ${3} ]  && return 1
test "$(md5sum ${1} | tr -s ' ' | cut -f1 -d' ')" = "$(zip_read ${2}.md5sum)"
return $?
}

md5sum_cmp_len()
{
[ ! -e ${1} ] && return 1
test "$(dd if=${1} count=1 bs=${3} 2>/dev/null | md5sum | tr -s ' ' | cut -f1 -d' ')" = "$(zip_read ${2}.md5sum)"
return $?
}

get_part_size()
{
SIZE="$(grep '\<'${FLASHDEV}${1}'$' /proc/partitions | tr -s ' ' | cut -f4 -d' ')"

if [ "${SIZE}" = "" ] ; then
	echo 0
else
	echo "${SIZE}"
fi
}

do_repart()
{
qdialog $(center 240 100) -title "Repartition" -fn "${qtoolfont}" -c -1 "&Reboot" -o "Repartioning not available..."

reboot -f
}

get_mtd_erase()
{
printf "%d\n" 0x$(cat /proc/mtd | grep "^mtd0:" | cut -f3 -d' ')
}

check_progress()
{
qprogress_c -s ${2} "${CHKMSG} ${1}"
}

check_progress_1()
{
echo -n x | check_progress "${1}" 1 >/dev/null
}

check_md5sum()
{
	local MD5SUM

	if [ "${2}" != "" ] ; then
		MD5SUM=$(zip_read ${1} | check_progress "${1}" "${2}" | md5sum | tr -s ' ' | cut -f1 -d' ')

		if [ "${MD5SUM}" != "$(zip_read ${1}.md5sum)" ] ; then
			panic "update image ${1} corrupted" 
		fi
	fi
}

check_uimage()
{
	local TMP_IMAGE=/tmp/image
	local MD5SUM

	if [ "${3}" != "" ] ; then
		MD5SUM=$(zip_read ${1} | check_progress "${1}" "${3}" | tee "${TMP_IMAGE}" | md5sum | tr -s ' ' | cut -f1 -d' ')

		if [ "${MD5SUM}" != "$(zip_read ${1}.md5sum)" ] ; then
			panic "update image ${1} corrupted" 
		fi

		/sbin/ui_verify ${TMP_IMAGE} ${2} >/dev/null
		
		ret=$?

		rm -f ${TMP_IMAGE}

		return $ret
	fi

	return 1
}

update_uimage()
{
	local TMP_IMAGE=/tmp/image

	zip_read ${1} |
	qprogress_c -s ${3} "Updating uimage ${2}..." >${TMP_IMAGE}

	UI_WRITE=$(/sbin/ui_write ${TMP_IMAGE} ${4})

	if  [ $? -ne 0 ] ; then
		panic "image update ${1} failed.
${UI_WRITE}" 
	fi

	if [ -z "${4}" ] ; then
		if ! /sbin/ui_verify ${TMP_IMAGE} =${UI_WRITE} ; then
			panic "image update ${1} verify failed."
		fi

		if ! /usr/sbin/senv ${2} ${UI_WRITE} ; then
			panic "environment update ${2} failed."
		fi
	fi

	rm -f ${TMP_IMAGE}
}

#chvt 4
export HOME=/tmp
export DISPLAY=:0
/usr/bin/xinit /usr/bin/sWM -- /usr/bin/X &
sleep 3
killall -0 Xorg || reboot -f

xset s off
xset -dpms

if ! mount -o ro $1 /mnt ; then
	panic "unable to access update device" 
fi

cd /tmp

PART0_SIZE=$(get_part_size 0)
PART1_SIZE=$(get_part_size 1)
PART2_SIZE=$(get_part_size 2)
PART3_SIZE=$(get_part_size 3)

if [ $(( $PART0_SIZE + $PART1_SIZE  + $PART2_SIZE + $PART3_SIZE )) -lt $MIN_FLASH_SIZE ] ; then
	panic "insufficent flash memory"
fi

LAYOUT="$(grep '\<'${FLASHDEV}'[0-9]\+$' /proc/partitions | tr -s ' ' | cut -f5 -d' ')"
if [ "$(echo ${LAYOUT})" != "${FLASHDEV}0 ${FLASHDEV}1 ${FLASHDEV}2 ${FLASHDEV}3" ] ; then
	do_repart 0
else
	if [ $PART2_SIZE -lt $MIN_PART2_SIZE ] ; then
		do_repart 2
	else
		if [ $PART3_SIZE -lt $MIN_PART3_SIZE ] ; then
			do_repart 3
		fi
	fi
fi

if ! mount -o rw -t yaffs /dev/rw /rw ; then
	panic "cannot access data parition"
fi

if [ "$(df /rw | tail -n 1 | tr -s ' ' | cut -f4 -d' ')" -lt 10240 ] ; then
	qdialog $(center 240 140) -title "$(title 3)" -fn "${qtoolfont}" -a -i -1 "&Reboot" -o "Software update failed.

At least 10 MB of free storage
space is requiered.

Please remove some user data."

	umount /mnt
	umount /rw

	reboot -f
fi

qprogress_c -k 2>/dev/null

ROOTFS_LEN=$(zip_length ${ROOTFS})
if [ "${ROOTFS_LEN}" = "" ] ; then
	panic "Cannot get update ${ROOTFS}"
fi

KERNEL_LEN=$(zip_length ${KERNEL_VOL})
if [ "${KERNEL_LEN}" = "" ] ; then
	panic "Cannot get update ${KERNEL}"
fi

ULOAD_LEN=$(zip_length ${U_LOAD})
FPGADATA_BB_LEN=$(zip_length ${FPGADATA_BB_VOL})
FPGADATA_RF_LEN=$(zip_length ${FPGADATA_RF_VOL})
U_BOOT_LEN=$(zip_length ${U_BOOT_VOL})
BOOTPIC_LINUX_LEN=$(zip_length ${BOOTPIC_LINUX_VOL})
BOOTPIC_UBOOT_LEN=$(zip_length ${BOOTPIC_UBOOT_VOL})
RECOVERY_LEN=$(zip_length ${RECOVERY_VOL})
OPTFS_LEN=$(zip_length ${OPTFS})
DATAFS_LEN=$(zip_length ${DATAFS})
FIXENV_LEN=$(zip_length ${FIXENV})
REMOVELIST_LEN=$(zip_length ${REMOVELIST})

SUM=$((
	${ULOAD_LEN:-0} +
	${KERNEL_LEN:-0} +
	${FPGADATA_BB_LEN:-0} +
	${FPGADATA_RF_LEN:-0} +
	${U_BOOT_LEN:-0} +
	${BOOTPIC_LINUX_LEN:-0} +
	${BOOTPIC_UBOOT_LEN:-0} +
	${RECOVERY_LEN:-0} +
	${OPTFS_LEN:-0} +
	${DATAFS_LEN:-0} +
	${FIXENV_LEN:-0} +
	${REMOVELIST_LEN:-0} +
	2
))

qprogress_s $(center 300 70) -title "$(title 1)" -fn "${qtoolfont}" -s ${SUM} "${CHKMSG}"

check_uimage ${FPGADATA_BB_VOL} fpgadata_bb_vol ${FPGADATA_BB_LEN} && FPGADATA_BB_LEN=""
check_uimage ${FPGADATA_RF_VOL} fpgadata_rf_vol ${FPGADATA_RF_LEN} && FPGADATA_RF_LEN=""
check_uimage ${U_BOOT_VOL} u_boot_vol ${U_BOOT_LEN} && U_BOOT_LEN=""
check_uimage ${BOOTPIC_LINUX_VOL} bootpic_linux_vol ${BOOTPIC_LINUX_LEN} && BOOTPIC_LINUX_LEN=""
check_uimage ${BOOTPIC_UBOOT_VOL} bootpic_uboot_vol ${BOOTPIC_UBOOT_LEN} && BOOTPIC_UBOOT_LEN=""
check_uimage ${RECOVERY_VOL} rescue_vol ${RECOVERY_LEN} && RECOVERY_LEN=""
if check_uimage ${KERNEL_VOL} kernel_vol ${KERNEL_LEN} ; then
	[ "${RECOVERY_LEN}" = "" ] && KERNEL_LEN=""
fi

if [ "${ULOAD_LEN}" != "" ] ; then
	check_md5sum ${U_LOAD} ${ULOAD_LEN}
	if md5sum_cmp_len /dev/mtd0 ${U_LOAD} ${ULOAD_LEN} ; then
		ULOAD_LEN=""
	fi
fi

check_progress_1 ${ROOTFS}
if md5sum_cmp /rw/${ROOTFS} ${ROOTFS} ${ROOTFS_LEN} ; then
	ROOTFS_LEN=""
fi

if [ "${OPTFS_LEN}" != "" ] ; then
	check_md5sum ${OPTFS} ${OPTFS_LEN}
fi

[ "${DATAFS_LEN}" != "" ] && check_md5sum ${DATAFS} ${DATAFS_LEN}
[ "${REMOVELIST_LEN}" != "" ] && check_md5sum ${REMOVELIST} ${REMOVELIST_LEN}
[ "${FIXENV_LEN}" != "" ] && check_md5sum ${FIXENV} ${FIXENV_LEN}

check_progress_1 "Done"
qprogress_c -k 2>/dev/null

SUM=$((
	${ULOAD_LEN:-0} +
	${KERNEL_LEN:-0} +
	${FPGADATA_BB_LEN:-0} +
	${FPGADATA_RF_LEN:-0} +
	${U_BOOT_LEN:-0} +
	${BOOTPIC_LINUX_LEN:-0} +
	${BOOTPIC_UBOOT_LEN:-0} +
	${RECOVERY_LEN:-0} +
	${ROOTFS_LEN:-0} +
	${OPTFS_LEN:-0} +
	${DATAFS_LEN:-0} +
	${FIXENV_LEN:-0} +
	${REMOVELIST_LEN:-0} +
	4
))

qprogress_s $(center 300 70) -title "$(title 2)" -fn "${qtoolfont}" -s ${SUM} "Performing update..."

rm -rf /rw/${SUBFS} /rw/devel /rw/testing /rw/opt /rw/startup /rw/var/cache/* 2>/dev/null

tar -cf ${SAVE} -C /rw/etc \
 bluetooth.pin \
 rs232.conf \
 security \
 sysctl.conf \
 systemsecpw.srspw \
 vncpasswd \
 network/interfaces \
 network/netbios\
 network/xitami \
 network/avahi \
 samba/smbpasswd \
 samba/workgroup.conf

{
grep -e "^${share_user}:" -v /etc/passwd
{
grep -e "^${share_user}:" /rw/etc/passwd
grep -e "^${share_user}:" /etc/passwd
} | head -n 1
} >/tmp/passwd

rm -rf /rw/etc

if [ -d /rw/var/user ] ; then
	{ rmdir /rw/var/user/share && rmdir /rw/var/user ; } 2>/dev/null ||
	mv /rw/var/user /rw/var/${board}/
fi
if [ "${board}" = "$(readlink /rw/var/user)" ] ; then
	rm /rw/var/user
fi
if [ ! -e /rw/var/user ] ; then
	ln -s "/var/${board}" /rw/var/user
fi

if [ "${ULOAD_LEN}" != "" ] ; then
	flash_erase /dev/mtd0 0 1
	zip_read ${U_LOAD} |
	qprogress_c -s ${ULOAD_LEN} "Updating Boot Loader..." |
	dd of=/dev/mtd0 bs=$(get_mtd_erase) count=1 conv=sync
	if ! md5sum_cmp_len /dev/mtd0 ${U_LOAD} ${ULOAD_LEN} ; then
		panic "Boot Loader Update failed..."
	fi
fi

{
	echo "reserved_vol"
	if [ "${KERNEL_LEN}" != "" ] ; then
		echo "bootmode bootrescue"
		echo "kernel_vol"
	fi
} | /usr/sbin/uenv

[ "${U_BOOT_LEN}" != "" ] && update_uimage ${U_BOOT_VOL} u_boot_vol ${U_BOOT_LEN}
[ "${FPGADATA_BB_LEN}" != "" ] && update_uimage ${FPGADATA_BB_VOL} fpgadata_bb_vol ${FPGADATA_BB_LEN}
[ "${FPGADATA_RF_LEN}" != "" ] && update_uimage ${FPGADATA_RF_VOL} fpgadata_rf_vol ${FPGADATA_RF_LEN}
[ "${BOOTPIC_LINUX_LEN}" != "" ] && update_uimage ${BOOTPIC_LINUX_VOL} bootpic_linux_vol ${BOOTPIC_LINUX_LEN}
[ "${BOOTPIC_UBOOT_LEN}" != "" ] && update_uimage ${BOOTPIC_UBOOT_VOL} bootpic_uboot_vol ${BOOTPIC_UBOOT_LEN}
[ "${RECOVERY_LEN}" != "" ] && update_uimage ${RECOVERY_VOL} rescue_vol ${RECOVERY_LEN}
[ "${KERNEL_LEN}" != "" ] && update_uimage ${KERNEL_VOL} kernel_vol ${KERNEL_LEN}

{
echo "bootmode mtd"
if [ "${FIXENV_LEN}" != "" ] ; then
	zip_read ${FIXENV} |
	qprogress_c -s ${FIXENV_LEN} "Updating U-Boot Environment..." 
fi
} | /usr/sbin/uenv

if [ "${REMOVELIST_LEN}" != "" ] ; then
	zip_read ${REMOVELIST} |
	qprogress_c -s ${REMOVELIST_LEN} "Remove obsolete files..." |
	while read l
	do
		eval rm -rf ${l}
	done
fi

rm -f /rw/${OPTFS} /rw/${OPTFS}.new

for i in /dev/loop*
do
	losetup -d $i 2>/dev/null
done

if [ "${ROOTFS_LEN}" != "" ] ; then
	(
	zip_read ${ROOTFS}
	cat /dev/zero
	) |
	qprogress_c -s ${ROOTFS_LEN} "Updating Rootfs..." >/rw/${ROOTFS}.new
	sync
	if ! md5sum_cmp /rw/${ROOTFS}.new ${ROOTFS} ${ROOTFS_LEN} ; then
		panic_space /rw/${ROOTFS}.new
	fi
	mv -f /rw/${ROOTFS}.new /rw/${ROOTFS}
	sync
fi

(
echo -n 1
(
cd /rw
cp -a /etc etc.new || panic_space /rw/etc.new
tar -xf ${SAVE} -C etc.new || panic_space /rw/etc.new
cp /tmp/passwd etc.new/passwd || panic_space /rw/etc.new
mv etc.new etc
) >&2
sync
echo -n 2
) |
qprogress_c -s 2 "Update settings..." >/dev/null

if [ "${OPTFS_LEN}" != "" ] ; then
	(
	zip_read ${OPTFS}
	cat /dev/zero
	) |
	qprogress_c -s ${OPTFS_LEN} "Updating Application..." >/rw/${OPTFS}.new
	sync
	if ! md5sum_cmp /rw/${OPTFS}.new ${OPTFS} ${OPTFS_LEN} ; then
		panic_space /rw/${OPTFS}.new
	fi
	mv -f /rw/${OPTFS}.new /rw/${OPTFS}
	sync
fi

if [ "${DATAFS_LEN}" != "" ] ; then
	(
	zip_read ${DATAFS}
	cat /dev/zero
	) |
	qprogress_c -s ${DATAFS_LEN} "Updating Application Data..." |
	tar -xz -C /rw
fi

(
echo -n 1
sync
sleep 2
echo -n 2
) |
qprogress_c -s 2 "Sync Filesystems..." >/dev/null

qprogress_c -k 2>/dev/null

if mount -o rw -t yaffs /dev/recovery /recovery ; then
	if [ ! -f "${RECOVERY}" ] ; then
		if [ ! -f "/recovery/${VERSIONINFO}" ] ; then
			qprogress $(center 300 70) -title "Init recovery system" -fn "${qtoolfont}" -f ${ZIPFILE} "Create recovery..." >${RECOVERY}.save

			if cmp ${ZIPFILE} ${RECOVERY}.save ; then
				mv ${RECOVERY}.save ${RECOVERY}
			else
				rm ${RECOVERY}.save
				if mount -o loop,ro "/rw/${OPTFS}" /opt ; then
					cp -f /opt/${board}/bin/${VERSIONINFO} /recovery
					umount /opt
				fi
			fi
		fi
	fi
	if ! unzip -o ${ZIPFILE} ${DATAFS} -d /recovery 2>/dev/null ; then
		rm /recovery/${DATAFS} 
	fi
	umount /recovery
fi

umount /rw
umount /mnt

qdialog $(center 240 140) -title "$(title 3)" -fn "${qtoolfont}" -a -i -1 "&Power Off" -o "Software update successful.

Please remove USB Stick
and press any key to
switch off instrument"

poweroff -f

